import hashlib, os
from typing import Any, Dict, Optional, Tuple

import torch

from ..version import __version__
from .generic import get_unique_timestamp, get_git_hash

# these are the base keys for the model dictionary to save
model_dict_full = {
    "epoch": 0,
    "model_state_dict": None,
    "optimizer_state_dict": None,
    "loss": None,
    "timestamp": None,
    "timestamp_hash": None,
    "git_hash": None,
    "version": None,
}

model_dict_required = {
    "model_state_dict": None,
    "optimizer_state_dict": None,
}

best_model_path_end = "_best.pth.tar"
latest_model_path_end = "_latest.pth.tar"
initial_model_path_end = "_initial.pth.tar"


def optimize_and_save_model(
    model: torch.nn.Module, params: dict, path: str, onnx_export: Optional[bool] = True
) -> None:
    """
    Perform post-training optimization and save it to a file.

    Args:
        model (torch.nn.Module): Trained torch model.
        params (dict): The parameter dictionary.
        path (str): The path to save the model dictionary to.
        onnx_export (Optional[bool]): Whether to export to ONNX and OpenVINO. Defaults to True.
    """
    # Check if ONNX export is enabled in the parameter dictionary
    onnx_export = params["model"].get("onnx_export", onnx_export)

    # Check for incompatible topologies and disable ONNX export
    # Customized imagenet_vgg no longer supported for ONNX export
    if onnx_export:
        architecture = params["model"]["architecture"]
        if architecture in ["sdnet", "brain_age"] or "imagenet_vgg" in architecture:
            onnx_export = False

    if not onnx_export:
        # Print a warning if ONNX export is disabled and not already warned
        if "onnx_print" not in params:
            print("WARNING: Current model is not supported by ONNX/OpenVINO!")
            params["onnx_print"] = True
        return
    else:
        try:
            print("Optimizing the best model.")
            num_channel = params["model"]["num_channels"]
            model_dimension = params["model"]["dimension"]
            input_shape = params["patch_size"]
            onnx_path = path
            if not onnx_path.endswith(".onnx"):
                onnx_path = onnx_path.replace("pth.tar", "onnx")

            if model_dimension == 2:
                dummy_input = torch.randn(
                    (1, num_channel, input_shape[0], input_shape[1])
                )
            else:
                dummy_input = torch.randn(
                    (1, num_channel, input_shape[0], input_shape[1], input_shape[2])
                )

            # Export the model to ONNX format
            with torch.no_grad():
                torch.onnx.export(
                    model.to("cpu"),
                    dummy_input.to("cpu"),
                    onnx_path,
                    opset_version=11,
                    export_params=True,
                    verbose=True,
                    input_names=["input"],
                    output_names=["output"],
                )
        except RuntimeWarning:
            print("WARNING: Cannot export to ONNX model.")
            return

        # Check if OpenVINO is present and try to convert the ONNX model
        openvino_present = False
        try:
            import openvino as ov
            from openvino.tools.mo import convert_model
            from openvino.runtime import get_version

            openvino_present = False
            # check for the correct openvino version to prevent inadvertent api breaks
            if "2023.0.1" in get_version():
                openvino_present = True
        except ImportError:
            print("WARNING: OpenVINO is not present.")

        if openvino_present:
            xml_path = onnx_path.replace("onnx", "xml")
            bin_path = onnx_path.replace("onnx", "bin")
            try:
                if model_dimension == 2:
                    ov_model = convert_model(
                        onnx_path,
                        input_shape=(1, num_channel, input_shape[0], input_shape[1]),
                    )
                else:
                    ov_model = convert_model(
                        onnx_path,
                        input_shape=(
                            1,
                            num_channel,
                            input_shape[0],
                            input_shape[1],
                            input_shape[2],
                        ),
                    )
                ov.runtime.serialize(ov_model, xml_path=xml_path, bin_path=bin_path)
            except Exception as e:
                print("WARNING: OpenVINO Model Optimizer IR conversion failed: " + e)


def save_model(
    model_dict: Dict[str, Any],
    model: torch.nn.Module,
    params: Dict[str, Any],
    path: str,
    onnx_export: Optional[bool] = True,
) -> None:
    """
    Save the model dictionary to a file.

    Args:
        model_dict (dict): Model dictionary to save.
        model (torch.nn.Module): Trained torch model.
        params (dict): The parameter dictionary.
        path (str): The path to save the model dictionary to.
        onnx_export (Optional[bool]): Whether to export to ONNX and OpenVINO. Defaults to True.
    """
    model_dict["timestamp"] = get_unique_timestamp()
    model_dict["timestamp_hash"] = hashlib.sha256(
        str(model_dict["timestamp"]).encode("utf-8")
    ).hexdigest()
    model_dict["version"] = __version__
    model_dict["parameters"] = params
    model_dict["git_hash"] = get_git_hash()

    torch.save(model_dict, path)

    # post-training optimization
    optimize_and_save_model(model, params, path, onnx_export=onnx_export)


def load_model(
    path: str, device: torch.device, full_sanity_check: Optional[bool] = True
) -> Dict[str, Any]:
    """
    Load a model dictionary from a file.

    Args:
        path (str): The path to save the model dictionary to.
        device (torch.device): The device to run the model on.
        full_sanity_check (Optional[bool]): Whether to perform a full sanity check. Defaults to True.

    Returns:
        dict: Model dictionary containing model parameters and metadata.
    """
    model_dict = torch.load(path, map_location=device)

    # check if the model dictionary is complete
    if full_sanity_check:
        incomplete_keys = [
            key for key in model_dict_full.keys() if key not in model_dict.keys()
        ]
        assert (
            len(incomplete_keys) == 0
        ), "Model dictionary is incomplete; the following keys are missing: " + str(
            incomplete_keys
        )

    # check if required keys are absent, and if so raise an error
    incomplete_required_keys = [
        key for key in model_dict_required.keys() if key not in model_dict.keys()
    ]
    assert (
        len(incomplete_required_keys) == 0
    ), "Model dictionary is incomplete; the following keys are missing: " + str(
        incomplete_required_keys
    )

    return model_dict


def load_ov_model(path: str, device: Optional[str] = "CPU") -> Tuple[Any, Any, Any]:
    """
    Load an OpenVINO IR model from an .xml file.

    Args:
        path (str): The path to the OpenVINO .xml file.
        device (Optional[str]): The device to run the model on, can be "CPU", "GPU" or "MULTI:CPU,GPU". Defaults to "CPU".

    Returns:
        Tuple[Any, Any, Any]: The compiled OpenVINO model, input layer name, and output layer name.
    """

    try:
        from openvino import runtime as ov
    except ImportError:
        raise ImportError("OpenVINO inference engine is not configured correctly.")

    core = ov.Core()
    if device.lower() == "cuda":
        device = "GPU"

    if device == "GPU":
        core.set_property({"CACHE_DIR": os.path.dirname(os.path.abspath(path))})

    model = core.read_model(model=path, weights=path.replace("xml", "bin"))
    compiled_model = core.compile_model(model=model, device_name=device.upper())
    input_layer = compiled_model.inputs
    output_layer = compiled_model.outputs

    return compiled_model, input_layer, output_layer



from .morphology import torch_morphological, fill_holes, cca

from .tensor import get_mapped_label


global_postprocessing_dict = {
    "fill_holes": fill_holes,
    "mapping": get_mapped_label,
    "morphology": torch_morphological,
    "cca": cca,
}

# append post_processing functions that are to be be applied after reverse one-hot encoding
postprocessing_after_reverse_one_hot_encoding = ["mapping"]



# adapted from https://github.com/kenshohara/3D-ResNets-PyTorch

import sys
import torch
import torch.nn as nn
import torch.nn.functional as F
from collections import OrderedDict

from .modelBase import ModelBase


class _DenseLayer(nn.Sequential):
    def __init__(self, num_input_features, growth_rate, bn_size, drop_rate, Norm, Conv):
        """
        Constructor for _DenseLayer class.

        Args:
            num_input_features (int): Number of input channels to the layer.
            growth_rate (int): Number of output channels of each convolution operation in the layer.
            bn_size (int): Factor to scale the number of intermediate channels between the 1x1 and 3x3 convolutions.
            drop_rate (float): Probability of an element to be zeroed in the Dropout layer.
            Norm (torch.nn.Module): A normalization module from torch.nn, such as BatchNorm2d or InstanceNorm2d.
            Conv (torch.nn.Module): A convolution module from torch.nn, such as Conv2d or ConvTranspose2d.
        """
        # Call the constructor of the parent class
        super().__init__()

        # Add a batch normalization layer followed by a ReLU activation function and a 1x1 convolution layer
        self.add_module("norm1", Norm(num_input_features))
        self.add_module("relu1", nn.ReLU(inplace=True))
        self.add_module(
            "conv1",
            Conv(
                num_input_features,
                bn_size * growth_rate,
                kernel_size=1,
                stride=1,
                bias=False,
            ),
        )

        # Add another batch normalization layer followed by a ReLU activation function and a 3x3 convolution layer
        self.add_module("norm2", Norm(bn_size * growth_rate))
        self.add_module("relu2", nn.ReLU(inplace=True))
        self.add_module(
            "conv2",
            Conv(
                bn_size * growth_rate,
                growth_rate,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
        )

        # Set the dropout rate
        self.drop_rate = drop_rate

    def forward(self, x):
        """
        Forward pass through the _DenseLayer.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Output tensor obtained after concatenating the input tensor and the new features obtained after the convolution operations and dropout.
        """
        # Perform forward pass through the layers
        new_features = super().forward(x)

        # If dropout rate is greater than 0, apply dropout to the new features
        if self.drop_rate > 0:
            new_features = F.dropout(
                new_features, p=self.drop_rate, training=self.training
            )

        # Concatenate the input tensor with the new features and return the result
        return torch.cat([x, new_features], 1)


class _DenseBlock(nn.Sequential):
    def __init__(
        self,
        num_layers,
        num_input_features,
        bn_size,
        growth_rate,
        drop_rate,
        norm,
        conv,
    ):
        """
        Constructor for _DenseBlock class.

        Args:
            num_layers (int): Number of dense layers to be added to the block.
            num_input_features (int): Number of input channels to the block.
            bn_size (int): Factor to scale the number of intermediate channels between the 1x1 and 3x3 convolutions in each dense layer.
            growth_rate (int): Number of output channels of each convolution operation in each dense layer.
            drop_rate (float): Probability of an element to be zeroed in the Dropout layer of each dense layer.
            norm (torch.nn.Module): A normalization module from torch.nn, such as BatchNorm2d or InstanceNorm2d, to be used in each dense layer.
            conv (torch.nn.Module): A convolution module from torch.nn, such as Conv2d or ConvTranspose2d, to be used in each dense layer.
        """
        # Call the constructor of the parent class
        super().__init__()

        # Add num_layers _DenseLayer objects to the block
        for i in range(num_layers):
            # Calculate the number of input features for the i-th dense layer
            num_input_features_i = num_input_features + i * growth_rate

            # Create an instance of _DenseLayer with the calculated number of input features and other parameters
            layer = _DenseLayer(
                num_input_features_i,
                growth_rate,
                bn_size,
                drop_rate,
                norm,
                conv,
            )

            # Add the _DenseLayer object to the block
            self.add_module("denselayer{}".format(i + 1), layer)


class _Transition(nn.Sequential):
    def __init__(self, num_input_features, num_output_features, Norm, Conv, AvgPool):
        super().__init__()
        self.add_module("norm", Norm(num_input_features))
        self.add_module("relu", nn.ReLU(inplace=True))
        self.add_module(
            "conv",
            Conv(
                num_input_features,
                num_output_features,
                kernel_size=1,
                stride=1,
                bias=False,
            ),
        )
        self.add_module("pool", AvgPool(kernel_size=2, stride=2))


class DenseNet(ModelBase):
    def __init__(
        self,
        parameters: dict,
        block_config=(6, 12, 24, 16),
    ):
        """
        Densenet-BC model class

        Args:
            block_config (list of 4 ints): how many layers in each pooling block
            parameters (dict): overall parameters dictionary; parameters specific for DenseNet:
                - growth_rate (int): how many filters to add each layer (k in paper)
                - num_init_features (int): the number of filters to learn in the first convolution layer
                - bn_size (int): multiplicative factor for number of bottle neck layers
                (i.e. bn_size * k features in the bottleneck layer)
                - drop_rate (float): dropout rate after each dense layer
                - num_classes (int): number of classification classes
                - final_convolution_layer (str): the final convolutional layer to use
                - norm_type (str): the normalization type to use
        """

        super(DenseNet, self).__init__(parameters)

        # defining some defaults
        if not ("num_init_features" in parameters):
            parameters["num_init_features"] = 64
        if not ("growth_rate" in parameters):
            parameters["growth_rate"] = 32
        if not ("bn_size" in parameters):
            parameters["bn_size"] = 4
        if not ("drop_rate" in parameters):
            parameters["drop_rate"] = 0
        if not ("conv1_t_stride" in parameters):
            parameters["conv1_t_stride"] = 1
        if not ("conv1_t_size" in parameters):
            parameters["conv1_t_size"] = 7
        if not ("no_max_pool" in parameters):
            parameters["no_max_pool"] = False
        if self.Norm is None:
            sys.stderr.write(
                "Warning: densenet is not defined without a normalization layer"
            )
            self.Norm = self.BatchNorm

        if self.n_dimensions == 2:
            self.output_size = (1, 1)
            self.conv_stride = (parameters["conv1_t_stride"], 2)
        elif self.n_dimensions == 3:
            self.output_size = (1, 1, 1)
            self.conv_stride = (parameters["conv1_t_stride"], 2, 2)
        else:
            sys.exit("Only 2D or 3D convolutions are supported.")

        # First convolution
        self.features = [
            (
                "conv1",
                self.Conv(
                    self.n_channels,
                    parameters["num_init_features"],
                    kernel_size=parameters["conv1_t_size"],
                    stride=self.conv_stride,
                    padding=parameters["conv1_t_size"] // 2,
                    bias=False,
                ),
            ),
            ("norm1", self.Norm(parameters["num_init_features"])),
            ("relu1", nn.ReLU(inplace=True)),
        ]
        if not parameters["no_max_pool"]:
            self.features.append(
                ("pool1", self.MaxPool(kernel_size=3, stride=2, padding=1))
            )
        self.features = nn.Sequential(OrderedDict(self.features))

        # Each denseblock
        num_features = parameters["num_init_features"]
        for i, num_layers in enumerate(block_config):
            block = _DenseBlock(
                num_layers=num_layers,
                num_input_features=num_features,
                bn_size=parameters["bn_size"],
                growth_rate=parameters["growth_rate"],
                drop_rate=parameters["drop_rate"],
                norm=self.Norm,
                conv=self.Conv,
            )
            self.features.add_module("denseblock{}".format(i + 1), block)
            num_features = num_features + num_layers * parameters["growth_rate"]
            if i != len(block_config) - 1:
                trans = _Transition(
                    num_input_features=num_features,
                    num_output_features=num_features // 2,
                    Norm=self.Norm,
                    Conv=self.Conv,
                    AvgPool=self.AvgPool,
                )
                self.features.add_module("transition{}".format(i + 1), trans)
                num_features = num_features // 2

        # Final batch norm
        self.features.add_module("norm5", self.Norm(num_features))

        for m in self.modules():
            if isinstance(m, self.Conv):
                m.weight = nn.init.kaiming_normal_(m.weight, mode="fan_out")
            elif isinstance(m, self.Norm):
                if m.weight is not None:
                    m.weight.data.fill_(1)
                if m.bias is not None:
                    m.bias.data.zero_()

        # Linear layer
        self.classifier = nn.Linear(num_features, self.n_classes)

        for m in self.modules():
            if isinstance(m, self.Conv):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, self.Norm):
                if m.weight is not None:
                    m.weight.data.fill_(1)
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        """
        Forward pass of the network

        Args:
            x (torch.Tensor) - the input tensor

        Returns:
            out (torch.Tensor) - the output tensor

        """
        # Pass the input tensor through the convolutional layers of the model
        features = self.features(x)

        # Apply a ReLU activation function to the feature maps
        out = F.relu(features, inplace=True)

        # Apply adaptive average pooling to the feature maps and flatten the resulting tensor
        out = self.AdaptiveAvgPool(self.output_size)(out).view(features.size(0), -1)

        # Pass the flattened tensor through the fully connected layers of the model
        out = self.classifier(out)

        # Apply the final convolutional operation, if specified
        if not self.final_convolution_layer is None:
            if self.final_convolution_layer == F.softmax:
                out = self.final_convolution_layer(out, dim=1)
            else:
                out = self.final_convolution_layer(out)

        # Return the output tensor
        return out


def densenet121(parameters):
    return DenseNet(
        parameters,
        block_config=(6, 12, 24, 16),
    )


def densenet169(parameters):
    return DenseNet(parameters, block_config=(6, 12, 32, 32))


def densenet201(parameters):
    return DenseNet(parameters, block_config=(6, 12, 48, 32))


def densenet264(parameters):
    return DenseNet(parameters, block_config=(6, 12, 64, 48))



from .modelBase import ModelBase
from GANDLF.models.seg_modules.out_conv import out_conv
import torch
import torch.nn as nn
from torch.nn import ModuleList
import numpy as np
import math


class _DeconvConvBlock(nn.Sequential):
    """
    A block consisting of a transposed convolutional layer, followed by a convolutional layer,
    a normalization layer, and a ReLU activation function. The block is defined as a sequential
    module in PyTorch, making it easy to stack multiple blocks together.

    Args:
        in_feats (int): The number of input features to the block.
        out_feats (int): The number of output features from the block.
        Norm (torch.nn.Module): The normalization layer to use (e.g. BatchNorm2d).
        Conv (torch.nn.Module): The convolutional layer to use (e.g. Conv2d).
        Deconv (torch.nn.Module): The transposed convolutional layer to use (e.g. ConvTranspose2d).
    """

    def __init__(self, in_feats, out_feats, Norm, Conv, Deconv):
        super().__init__()

        self.add_module(
            "deconv",
            Deconv(
                in_channels=in_feats,
                out_channels=out_feats,
                kernel_size=2,
                stride=2,
                padding=0,
                output_padding=0,
            ),
        )

        self.add_module(
            "conv",
            Conv(
                in_channels=out_feats,
                out_channels=out_feats,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
        )

        self.add_module("norm", Norm(out_feats))
        self.add_module("relu", nn.ReLU(inplace=True))

    def forward(self, x):
        """
        Forward pass through the block.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            x (torch.Tensor): The output tensor.
        """

        # Apply deconv -> conv -> norm -> relu
        x = self.deconv(x)
        x = self.conv(x)
        x = self.norm(x)
        x = self.relu(x)

        return x


class _ConvBlock(nn.Sequential):
    """
    A block consisting of a convolutional layer followed by batch normalization and ReLU activation.

    Args:
        in_feats (int): Number of input features.
        out_feats (int): Number of output features.
        Norm (nn.Module): A normalization layer.
        Conv (nn.Module): A convolutional layer.
    """

    def __init__(self, in_feats, out_feats, Norm, Conv):
        """
        Initializes the ConvBlock.

        Args:
            in_feats (int): Number of input features.
            out_feats (int): Number of output features.
            Norm (nn.Module): A normalization layer.
            Conv (nn.Module): A convolutional layer.
        """
        super().__init__()

        self.add_module(
            "conv",
            Conv(
                in_channels=in_feats,
                out_channels=out_feats,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
        )

        self.add_module("norm", Norm(out_feats))
        self.add_module("relu", nn.ReLU(inplace=True))

    def forward(self, x):
        """
        Performs a forward pass of the ConvBlock.

        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, in_feats, height, width).

        Returns:
            torch.Tensor: Output tensor of shape (batch_size, out_feats, height, width).
        """

        # Apply conv -> norm -> relu
        x = self.conv(x)
        x = self.norm(x)
        x = self.relu(x)

        return x


class _UpsampleBlock(nn.Sequential):
    def __init__(self, in_feats, Norm, Conv, Deconv):
        """
        Args:
            in_feats (int): Number of input channels.
            Norm (nn.Module): Normalization layer constructor.
            Conv (nn.Module): Convolutional layer constructor.
            Deconv (nn.Module): Transposed convolutional layer constructor.
        """
        super().__init__()

        self.add_module(
            "conv1",
            Conv(
                in_channels=in_feats,
                out_channels=in_feats // 2,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
        )

        self.add_module("norm1", Norm(in_feats // 2))
        self.add_module("relu1", nn.ReLU(inplace=True))

        self.add_module(
            "conv2",
            Conv(
                in_channels=in_feats // 2,
                out_channels=in_feats // 2,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
        )

        self.add_module("norm2", Norm(in_feats // 2))
        self.add_module("relu2", nn.ReLU(inplace=True))

        self.add_module(
            "deconv",
            Deconv(
                in_channels=in_feats // 2,
                out_channels=in_feats // 4,
                kernel_size=2,
                stride=2,
                padding=0,
                output_padding=0,
            ),
        )

    def forward(self, x):
        """
        Performs a forward pass through the upsampling block.

        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, in_feats, height, width).

        Returns:
            torch.Tensor: Output tensor of shape (batch_size, in_feats//4, 2*height, 2*width).
        """

        # Apply conv1 -> norm1 -> relu1-> relu2
        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu1(x)

        # Apply conv2 -> norm2 -> relu2
        x = self.conv2(x)
        x = self.norm2(x)
        x = self.relu2(x)

        # Apply deconv
        x = self.deconv(x)

        return x


class _MLP(nn.Sequential):
    """
    Multi-layer perceptron module that applies two linear transformations with GELU activations.

    Args:
        in_feats (int): Number of input features.
        out_feats (int): Number of output features for the first linear transformation.
        Norm (nn.Module): Normalization module, e.g. LayerNorm.

    """

    def __init__(self, in_feats, out_feats):  # Which normalization module to use?
        super().__init__()

        self.add_module("norm", nn.LayerNorm(in_feats))

        self.add_module("linear1", nn.Linear(in_feats, out_feats))
        self.add_module("gelu1", nn.GELU())

        self.add_module("linear2", nn.Linear(out_feats, in_feats))
        self.add_module("gelu2", nn.GELU())

    def forward(self, x):
        """
        Applies the MLP module to the input tensor.

        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, in_feats).

        Returns:
            torch.Tensor: The output tensor of shape (batch_size, in_feats).
        """

        # Apply norm -> linear1 -> gelu1
        x = self.norm(x)
        x = self.linear1(x)
        x = self.gelu1(x)

        # Apply linear2 -> gelu2
        x = self.linear2(x)
        x = self.gelu2(x)

        return x


class _MSA(nn.Module):
    """
    Multi-Head Self-Attention module.

    Args:
        embed_size (int): The size of the input feature embedding.
        num_heads (int): The number of attention heads to use.

    Attributes:
        num_heads (int): The number of attention heads used.
        query_size (int): The size of each query vector.
        all_heads (int): The total size of all query, key, and value vectors.

        query (nn.Linear): Linear layer to project input features into query vectors.
        key (nn.Linear): Linear layer to project input features into key vectors.
        value (nn.Linear): Linear layer to project input features into value vectors.
        out (nn.Linear): Linear layer to project concatenated attention head outputs.

        softmax (nn.Softmax): Softmax function used to compute attention weights.

    Methods:
        reshape(x): Reshapes the input tensor to enable batch-wise matrix multiplication.
        forward(x): Computes multi-head self-attention on input tensor.

    Example:
        >>> msa = _MSA(embed_size=256, num_heads=8)
        >>> x = torch.randn(4, 16, 256)  # batch size = 4, sequence length = 16, embedding size = 256
        >>> output = msa(x)  # output shape: (4, 16, 256)
    """

    def __init__(self, embed_size, num_heads):
        super().__init__()

        # Initialize parameters
        self.num_heads = num_heads
        self.query_size = int(embed_size / self.num_heads)
        self.all_heads = self.query_size * self.num_heads  # should equal embed_size

        # Initialize layers
        self.query = nn.Linear(embed_size, self.all_heads)
        self.key = nn.Linear(embed_size, self.all_heads)
        self.value = nn.Linear(embed_size, self.all_heads)

        self.out = nn.Linear(self.all_heads, self.all_heads)

        self.softmax = nn.Softmax(dim=-1)

    def reshape(self, x):
        """
        Reshapes the input tensor to enable batch-wise matrix multiplication.

        Args:
            x (torch.Tensor): The input tensor to reshape.

        Returns:
            The reshaped tensor.
        """
        x_shape = list(x.size()[:-1]) + [self.num_heads, self.query_size]
        x = x.view(*x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(self, x):
        """
        Computes multi-head self-attention on input tensor.

        Args:
            x (torch.Tensor): The input tensor of shape `(batch_size, sequence_length, embed_size)`.

        Returns:
            The output tensor of shape `(batch_size, sequence_length, embed_size)`.
        """
        query = self.reshape(self.query(x))
        key = self.reshape(self.key(x))
        value = self.reshape(self.value(x))

        attention_weights = self.softmax(
            torch.matmul(query, key.transpose(-1, -2)) / math.sqrt(self.query_size)
        )

        self_attention = torch.matmul(attention_weights, value)
        self_attention = self_attention.permute(0, 2, 1, 3).contiguous()
        self_attention = self_attention.view(
            list(self_attention.size()[:-2]) + [self.all_heads]
        )  # reshape

        msa = self.out(self_attention)

        return msa


class _Embedding(nn.Module):
    def __init__(self, img_size, patch_size, in_feats, embed_size, Conv):
        """
        A module that creates embeddings from an image tensor.

        Args:
            img_size (tuple[int]): The size of the input image tensor (H, W, C).
            patch_size (int): The size of each square patch that the image is divided into.
            in_feats (int): The number of input channels in the image tensor.
            embed_size (int): The size of the output embedding vector.
            Conv (nn.Module): The convolutional module to use for extracting patches.
        """
        super().__init__()

        # Calculate the number of patches in the image
        self.n_patches = int(np.prod([i / patch_size for i in img_size]))

        # Create a learnable parameter for the positional embedding
        self.pos_embed = nn.Parameter(torch.zeros(1, self.n_patches, embed_size))

        # Create a convolutional module to extract patches from the input image
        self.patch_embed = Conv(
            in_channels=in_feats,
            out_channels=embed_size,
            kernel_size=patch_size,
            stride=patch_size,
        )

    def forward(self, x):
        """
        Forward pass for the _Embedding module.

        Args:
            x (torch.Tensor): An input image tensor of shape (B, C, H, W).

        Returns:
            torch.Tensor: An output tensor of shape (B, n_patches, embed_size).
        """
        # Extract patches from the input image
        x = self.patch_embed(x)

        # Flatten the patches into a 2D matrix
        x = x.flatten(2)

        # Transpose the matrix so that patches are in the first dimension
        x = x.transpose(-1, -2)

        # Add the positional embedding to the flattened patches
        x = x + self.pos_embed

        return x


class _TransformerLayer(nn.Module):
    def __init__(self, embed_size, num_heads):
        """
        A module that implements a single transformer layer.

        Args:
            embed_size (int): The size of the embedding vector.
            num_heads (int): The number of heads to use in the multi-head self-attention module.
        """
        super().__init__()

        # Create normalization modules and a multi-head self-attention module
        self.norm1 = nn.LayerNorm(embed_size)
        self.msa = _MSA(embed_size, num_heads)

        # Create normalization modules and an MLP
        self.norm2 = nn.LayerNorm(embed_size)

        # Dev note: it should be out_feats=mlp_dim, but we have out_feats=num_heads
        # Also, the Norm parameter is not used in the MLP module
        # So this needs to be looked at later, but for now, it works
        self.mlp = _MLP(in_feats=embed_size, out_feats=num_heads)

    def forward(self, x):
        """
        Forward pass for the _TransformerLayer module.

        Args:
            x (torch.Tensor): An input tensor of shape (B, n_patches, embed_size).

        Returns:
            torch.Tensor: An output tensor of the same shape as the input.
        """
        # Normalize the input and apply multi-head self-attention
        y = self.norm1(x)
        y = self.msa(y)

        # Add the input to the output of the self-attention module
        x = x + y

        # Normalize the output of the self
        y = self.norm2(x)
        y = self.mlp(y)

        # Add the input to the output of the MLP
        x = x + y

        return x


class _Transformer(nn.Sequential):
    """
    A transformer module that consists of an embedding layer followed by a series of transformer layers.

    Args:
        img_size (tuple): The dimensions of the input image (height, width, depth).
        patch_size (int): The size of the patches to be extracted from the input image.
        in_feats (int): The number of input features.
        embed_size (int): The size of the embedding.
        num_heads (int): The number of attention heads to use in the multi-head attention layer.
        num_layers (int): The number of transformer layers to use.
        out_layers (list): A list of indices indicating which transformer layers should output their results.
        Norm (nn.Module): A normalization module to use for the transformer layers.

    Attributes:
        out_layers (list): A list of indices indicating which transformer layers should output their results.
        num_layers (int): The number of transformer layers to use.
        embed (_Embedding): The embedding layer.
        layers (ModuleList): A list of _TransformerLayer objects.

    Methods:
        forward(x): Processes the input through the transformer and returns the output.

    Example:
        transformer = _Transformer(
            img_size=(224, 224, 3),
            patch_size=16,
            in_feats=3,
            embed_size=256,
            num_heads=8,
            mlp_dim=2048,
            num_layers=12,
            out_layers=[5, 11],
            Conv=nn.Conv2d,
            Norm=nn.LayerNorm
        )
        output = transformer(input_tensor)
    """

    def __init__(
        self,
        img_size,
        patch_size,
        in_feats,
        embed_size,
        num_heads,
        mlp_dim,
        num_layers,
        out_layers,
        Conv,
        Norm,
    ):
        super().__init__()
        self.out_layers = out_layers
        self.num_layers = num_layers
        self.embed = _Embedding(img_size, patch_size, in_feats, embed_size, Conv)
        self.layers = ModuleList([])

        for _ in range(0, num_layers):
            layer = _TransformerLayer(
                embed_size,
                num_heads,
            )
            self.layers.append(layer)

    def forward(self, x):
        """
        Processes the input through the transformer and returns the output.

        Args:
            x (tensor): The input tensor.

        Returns:
            out (list): A list of tensors representing the output of the specified transformer layers.
        """
        out = []
        x = self.embed(x)

        # Note : Lists are highly inefficient when dealing with tensor operations and must be used carefully
        for i in range(0, self.num_layers):
            x = self.layers[i](x)
            if i in self.out_layers:
                out.append(x)

        return out


def checkImgSize(img_size, number=4):
    """
    Checks if the input image size is greater than or equal to 2^number in each dimension.
    If it is, it returns the specified number. Otherwise, it returns the minimum log base 2
    of the input image size.

    Args:
        img_size (tuple of ints): size of the input image tensor
        number (int): a number that the input image size is checked against. Default value is 4.

    Returns:
        int : An integer, either the specified number or the minimum log base 2 of the input image size.
    """
    if all([x >= 2**number for x in img_size]):
        return number
    else:
        return int(np.min(np.floor(np.log2(img_size))))


class unetr(ModelBase):
    """
    This is the U-NetR architecture : https://arxiv.org/abs/2103.10504.
    The 'residualConnections' flag controls residual connections The Downsampling, Encoding, Decoding modules are defined in the seg_modules file.
    These smaller modules are basically defined by 2 parameters, the input channels (filters) and the output channels (filters),
    and some other hyperparameters, which remain constant all the modules. For more details on the smaller modules please have a look at the seg_modules file.
    """

    def __init__(
        self,
        parameters: dict,
    ):
        """
        Initializes an instance of the `unetr` class.

        Args:
            parameters (dict): A dictionary containing the model parameters.
        """
        super(unetr, self).__init__(parameters)

        # initialize defaults if not found
        parameters["model"]["inner_patch_size"] = parameters["model"].get(
            "inner_patch_size", parameters["patch_size"][0]
        )
        parameters["model"]["num_heads"] = parameters["model"].get("num_heads", 12)
        parameters["model"]["embed_dim"] = parameters["model"].get("embed_dim", 768)

        self.patch_size = parameters["model"]["inner_patch_size"]

        assert np.ceil(np.log2(parameters["model"]["inner_patch_size"])) == np.floor(
            np.log2(parameters["model"]["inner_patch_size"])
        ), "The inner patch size must be a power of 2."

        self.depth = int(np.log2(self.patch_size))
        parameters["model"]["depth"] = self.depth

        _ = self.model_depth_check(parameters)

        if self.n_dimensions == 2:
            self.img_size = parameters["patch_size"][0:2]
        elif self.n_dimensions == 3:
            self.img_size = parameters["patch_size"]

        self.num_layers = 3 * self.depth  # number of transformer layers
        self.out_layers = np.arange(2, self.num_layers, 3)

        self.num_heads = parameters["model"]["num_heads"]
        self.embed_size = parameters["model"]["embed_dim"]

        self.patch_dim = [i // self.patch_size for i in self.img_size]

        assert (
            self.embed_size % self.num_heads == 0
        ), "The embedding dimension must be divisible by the number of self-attention heads"

        assert all(
            [i % self.patch_size == 0 for i in self.img_size]
        ), "The image size is not divisible by the patch size in at least 1 dimension. UNETR is not defined in this case."

        assert all(
            [self.patch_size <= i for i in self.img_size]
        ), "The inner patch size must be smaller than the input image."

        self.transformer = _Transformer(
            img_size=self.img_size,
            patch_size=self.patch_size,
            in_feats=self.n_channels,
            embed_size=self.embed_size,
            num_heads=self.num_heads,
            mlp_dim=2048,
            num_layers=self.num_layers,
            out_layers=self.out_layers,
            Conv=self.Conv,
            Norm=self.Norm,
        )

        self.upsampling = ModuleList([])
        self.convs = ModuleList([])

        for i in range(0, self.depth - 1):
            # add deconv blocks
            tempconvs = nn.Sequential()
            tempconvs.add_module(
                "conv0",
                _DeconvConvBlock(
                    self.embed_size,
                    32 * 2**self.depth,
                    self.Norm,
                    self.Conv,
                    self.ConvTranspose,
                ),
            )

            for j in range(self.depth - 2, i, -1):
                tempconvs.add_module(
                    "conv%d" % j,
                    _DeconvConvBlock(
                        128 * 2**j,
                        128 * 2 ** (j - 1),
                        self.Norm,
                        self.Conv,
                        self.ConvTranspose,
                    ),
                )

            self.convs.append(tempconvs)

            # add upsampling
            self.upsampling.append(
                _UpsampleBlock(
                    128 * 2 ** (i + 1), self.Norm, self.Conv, self.ConvTranspose
                )
            )

        # add upsampling for transformer output (no convs)
        self.upsampling.append(
            self.ConvTranspose(
                in_channels=self.embed_size,
                out_channels=32 * 2**self.depth,
                kernel_size=2,
                stride=2,
                padding=0,
                output_padding=0,
            )
        )

        self.input_conv = nn.Sequential()
        self.input_conv.add_module(
            "conv1", _ConvBlock(self.n_channels, 32, self.Norm, self.Conv)
        )
        self.input_conv.add_module("conv2", _ConvBlock(32, 64, self.Norm, self.Conv))

        # Dev note : Need to update the hard-coded number of channel modules here
        self.output_conv = nn.Sequential()
        self.output_conv.add_module("conv1", _ConvBlock(128, 64, self.Norm, self.Conv))
        self.output_conv.add_module("conv2", _ConvBlock(64, 64, self.Norm, self.Conv))
        self.output_conv.add_module(
            "conv3",
            out_conv(
                64,
                self.n_classes,
                conv_kwargs={
                    "kernel_size": 1,
                    "stride": 1,
                    "padding": 0,
                    "bias": False,
                },
                norm=self.Norm,
                conv=self.Conv,
                final_convolution_layer=self.final_convolution_layer,
                sigmoid_input_multiplier=self.sigmoid_input_multiplier,
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the U-Net model.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The output tensor.
        """
        # Perform transformer encoding of input tensor
        transformer_out = self.transformer(x)

        # Perform upsampling on last transformer output and concatenate with previous outputs
        y = self.upsampling[-1](
            transformer_out[-1]
            .transpose(-1, -2)
            .view(-1, self.embed_size, *self.patch_dim)
        )  # z12

        for i in range(len(self.convs) - 1, -1, -1):
            # Perform convolution on transformer output and concatenate with previous outputs
            zi = (
                transformer_out[i]
                .transpose(-1, -2)
                .view(-1, self.embed_size, *self.patch_dim)
            )

            zi = self.convs[i](zi)

            # Perform convolution on concatenated output
            zicat = torch.cat([zi, y], dim=1)

            # Perform upsampling on concatenated output
            y = self.upsampling[i](zicat)

        # Perform convolution on input tensor and concatenate with final output
        x = self.input_conv(x)
        x = torch.cat([x, y], dim=1)
        x = self.output_conv(x)

        return x



from typing import Optional, Union
import os
from pathlib import Path
import numpy as np

import pandas
import torch
import torchio
from torchio.transforms import Pad
import SimpleITK as sitk
from tqdm import tqdm

from GANDLF.utils import (
    perform_sanity_check_on_subject,
    resize_image,
    get_filename_extension_sanitized,
    get_correct_padding_size,
)
from .preprocessing import get_transforms_for_preprocessing
from .augmentation import global_augs_dict

global_sampler_dict = {
    "uniform": torchio.data.UniformSampler,
    "uniformsampler": torchio.data.UniformSampler,
    "uniformsample": torchio.data.UniformSampler,
    "label": torchio.data.LabelSampler,
    "labelsampler": torchio.data.LabelSampler,
    "labelsample": torchio.data.LabelSampler,
    "weighted": torchio.data.WeightedSampler,
    "weightedsampler": torchio.data.WeightedSampler,
    "weightedsample": torchio.data.WeightedSampler,
}


# This function takes in a dataframe, with some other parameters and returns the dataloader
def ImagesFromDataFrame(
    dataframe: pandas.DataFrame,
    parameters: dict,
    train: bool,
    apply_zero_crop: Optional[bool] = False,
    loader_type: Optional[str] = None,
) -> Union[torchio.SubjectsDataset, torchio.Queue]:
    """
    Reads the pandas dataframe and gives the dataloader to use for training/validation/testing.

    Args:
        dataframe (pandas.DataFrame): The main input dataframe which is calculated after splitting the data CSV.
        parameters (dict): The parameters dictionary.
        train (bool): If the dataloader is for training or not.
        apply_zero_crop (Optional[bool], optional): Whether to apply zero crop or not. Defaults to False.
        loader_type (Optional[str], optional): The type of loader. Defaults to None.

    Returns:
        Union[torchio.SubjectsDataset, torchio.Queue]: The dataloader queue for validation/testing (where patching and data augmentation is not required) or the subjects dataset for training.
    """
    loader_type = loader_type if loader_type is not None else ""
    # store in previous variable names
    patch_size = parameters["patch_size"]
    headers = parameters["headers"]
    q_max_length = parameters["q_max_length"]
    q_samples_per_volume = parameters["q_samples_per_volume"]
    q_num_workers = parameters["q_num_workers"]
    q_verbose = parameters["q_verbose"]
    augmentations = parameters["data_augmentation"]
    preprocessing = parameters["data_preprocessing"]
    in_memory = parameters["in_memory"]
    sampler = parameters["patch_sampler"]

    # Finding the dimension of the dataframe for computational purposes later
    num_row, num_col = dataframe.shape
    # changing the column indices to make it easier
    dataframe.columns = range(0, num_col)
    dataframe.index = range(0, num_row)
    # This list will later contain the list of subjects
    subjects_list = []
    subjects_with_error = []

    channelHeaders = headers["channelHeaders"]
    labelHeader = headers["labelHeader"]
    predictionHeaders = headers["predictionHeaders"]
    subjectIDHeader = headers["subjectIDHeader"]

    resize_images_flag = False
    # if resize has been defined but resample is not (or is none)
    if not (preprocessing is None):
        for key in preprocessing.keys():
            # check for different resizing keys
            if key in ["resize", "resize_image", "resize_images"]:
                if not (preprocessing[key] is None):
                    resize_images_flag = True
                    preprocessing["resize_image"] = preprocessing[key]
                    break

    # helper function to save the resized images
    def _save_resized_images(
        resized_image, output_dir, subject_id, channel_str, loader_type, extension
    ):
        """
        Helper function to save the resized images

        Args:
            resized_image (sitk.Image): The resized image.
            output_dir (str): The output directory.
            subject_id (str): The subject ID.
            channel_str (str): The channel string.
            loader_type (str): The loader type.
            extension (str): The extension of the image.
        """
        # save img_resized to disk
        save_dir_for_resized_images = os.path.join(
            output_dir, loader_type + "_resized_images"
        )
        Path(save_dir_for_resized_images).mkdir(parents=True, exist_ok=True)
        save_path = os.path.join(
            save_dir_for_resized_images,
            subject_id + "_" + channel_str + "_resized" + extension,
        )
        if not os.path.isfile(save_path):
            sitk.WriteImage(resized_image, save_path)

    # iterating through the dataframe
    for patient in tqdm(
        range(num_row), desc="Constructing queue for " + loader_type + " data"
    ):
        # We need this dict for storing the meta data for each subject
        # such as different image modalities, labels, any other data
        subject_dict = {}
        subject_dict["subject_id"] = str(dataframe[subjectIDHeader][patient])
        skip_subject = False
        # iterating through the channels/modalities/timepoints of the subject
        for channel in channelHeaders:
            # sanity check for malformed csv
            if not os.path.isfile(str(dataframe[channel][patient])):
                skip_subject = True

            subject_dict[str(channel)] = torchio.ScalarImage(
                dataframe[channel][patient]
            )

            # store image spacing information if not present
            if "spacing" not in subject_dict:
                file_reader = sitk.ImageFileReader()
                file_reader.SetFileName(str(dataframe[channel][patient]))
                file_reader.ReadImageInformation()
                subject_dict["spacing"] = torch.Tensor(file_reader.GetSpacing())

            # if resize_image is requested, the perform per-image resize with appropriate interpolator
            if resize_images_flag:
                img_resized = resize_image(
                    subject_dict[str(channel)].as_sitk(), preprocessing["resize_image"]
                )
                if parameters["memory_save_mode"]:
                    _save_resized_images(
                        img_resized,
                        parameters["output_dir"],
                        subject_dict["subject_id"],
                        str(channel),
                        loader_type,
                        get_filename_extension_sanitized(
                            str(dataframe[channel][patient])
                        ),
                    )
                else:
                    # always ensure resized image spacing is used
                    subject_dict["spacing"] = torch.Tensor(img_resized.GetSpacing())
                    subject_dict[str(channel)] = torchio.ScalarImage.from_sitk(
                        img_resized
                    )

        # # for regression -- this logic needs to be thought through
        # if predictionHeaders:
        #     # get the mask
        #     if (subject_dict['label'] is None) and (class_list is not None):
        #         sys.exit('The \'class_list\' parameter has been defined but a label file is not present for patient: ', patient)

        if labelHeader is not None:
            if not os.path.isfile(str(dataframe[labelHeader][patient])):
                skip_subject = True

            subject_dict["label"] = torchio.LabelMap(dataframe[labelHeader][patient])
            subject_dict["path_to_metadata"] = str(dataframe[labelHeader][patient])

            # if resize is requested, the perform per-image resize with appropriate interpolator
            if resize_images_flag:
                img_resized = resize_image(
                    subject_dict["label"].as_sitk(),
                    preprocessing["resize_image"],
                    sitk.sitkNearestNeighbor,
                )
                if parameters["memory_save_mode"]:
                    _save_resized_images(
                        img_resized,
                        parameters["output_dir"],
                        subject_dict["subject_id"],
                        "label",
                        loader_type,
                        get_filename_extension_sanitized(
                            str(dataframe[channel][patient])
                        ),
                    )
                else:
                    subject_dict["label"] = torchio.LabelMap.from_sitk(img_resized)

        else:
            subject_dict["label"] = "NA"
            subject_dict["path_to_metadata"] = str(dataframe[channel][patient])

        # iterating through the values to predict of the subject
        valueCounter = 0
        for values in predictionHeaders:
            # assigning the dict key to the channel
            subject_dict["value_" + str(valueCounter)] = np.array(
                dataframe[values][patient]
            )
            valueCounter += 1

        # skip subject the condition was tripped
        if not skip_subject:
            # Initializing the subject object using the dict
            subject = torchio.Subject(subject_dict)
            # https://github.com/fepegar/torchio/discussions/587#discussioncomment-928834
            # this is causing memory usage to explode, see https://github.com/mlcommons/GaNDLF/issues/128
            if parameters["verbose"]:
                print(
                    "Checking consistency of images in subject '"
                    + subject["subject_id"]
                    + "'"
                )
            try:
                perform_sanity_check_on_subject(subject, parameters)
            except Exception as exception:
                subjects_with_error.append(subject["subject_id"])
                print(
                    "Subject '"
                    + subject["subject_id"]
                    + "' could not be loaded due to the following exception: {}".format(
                        type(exception).__name__
                    )
                    + "; message: {}".format(exception)
                )

            # # padding image, but only for label sampler, because we don't want to pad for uniform
            if sampler["enable_padding"]:
                psize_pad = get_correct_padding_size(
                    patch_size, parameters["model"]["dimension"]
                )
                padder = Pad(psize_pad, padding_mode=sampler["padding_mode"])
                subject = padder(subject)

            # load subject into memory: https://github.com/fepegar/torchio/discussions/568#discussioncomment-859027
            if in_memory:
                subject.load()

            # Appending this subject to the list of subjects
            subjects_list.append(subject)

    assert (
        subjects_with_error is not None
    ), f"The following subjects could not be loaded, please recheck or remove and retry: {subjects_with_error}"

    transformations_list = []

    # augmentations are applied to the training set only
    if train and not (augmentations is None):
        for aug in augmentations:
            aug_lower = aug.lower()
            if aug_lower in global_augs_dict:
                transformations_list.append(
                    global_augs_dict[aug_lower](augmentations[aug])
                )

    transform = get_transforms_for_preprocessing(
        parameters, transformations_list, train, apply_zero_crop
    )

    subjects_dataset = torchio.SubjectsDataset(subjects_list, transform=transform)
    if not train:
        return subjects_dataset

    # initialize the sampler
    sampler_obj = global_sampler_dict[sampler["type"]](patch_size)
    if sampler["type"] in ("weighted", "weightedsampler", "weightedsample"):
        sampler_obj = global_sampler_dict[sampler["type"]](
            patch_size, probability_map="label"
        )
    elif sampler["type"] in ("label", "labelsampler", "labelsample"):
        # if biased sampling is detected, then we need to pass the class probabilities
        if train and sampler["biased_sampling"]:
            # initialize the class probabilities dict
            label_probabilities = {}
            if "sampling_weights" in parameters:
                for class_index in parameters["sampling_weights"]:
                    label_probabilities[class_index] = parameters["sampling_weights"][
                        class_index
                    ]
            sampler_obj = global_sampler_dict[sampler["type"]](
                patch_size, label_probabilities=label_probabilities
            )
    # all of these need to be read from model.yaml
    patches_queue = torchio.Queue(
        subjects_dataset,
        max_length=q_max_length,
        samples_per_volume=q_samples_per_volume,
        sampler=sampler_obj,
        num_workers=q_num_workers,
        shuffle_subjects=True,
        shuffle_patches=True,
        verbose=q_verbose,
    )
    return patches_queue
